import unittest
from io import BytesIO

from rethink import models, const
from . import utils


class VerificationTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        models.database.drop_all()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def test_verification_img(self):
        token, data = models.verification.random_captcha()
        self.assertIsInstance(token, str)
        self.assertIn("img", data)
        self.assertIsInstance(data["img"], BytesIO)
        self.assertNotIn("sound", data)

    def test_verify_img(self):
        token, data = models.verification.random_captcha()
        decoded = models.utils.jwt_decode(token)
        code = models.verification.verify_captcha(token=token, code_str=decoded["code"])
        self.assertEqual(const.Code.OK, code)

        code = models.verification.verify_captcha(token=token, code_str="1234")
        self.assertEqual(const.Code.CAPTCHA_ERROR, code)



import datetime
from typing import List, Optional, Tuple

from bson import ObjectId
from bson.tz_util import utc

from rethink import const, config
from . import tps, utils
from .database import COLL


def add(
        account: str,
        source: int,
        email: str,
        hashed: str,
        nickname: str,
        avatar: str,
        language: str,
) -> Tuple[str, const.Code]:
    if COLL.users.find_one({"account": account, "source": source}) is not None:
        return "", const.Code.EMAIL_OCCUPIED
    oid = ObjectId()
    # assert language in const.Language
    if not const.Language.is_valid(language):
        language = const.Language.EN.value
    data: tps.UserMeta = {
        "_id": oid,
        "id": utils.short_uuid(),
        "account": account,
        "source": source,
        "email": email,
        "hashed": hashed,
        "avatar": str(avatar),
        "disabled": False,
        "nickname": nickname,
        "modifiedAt": oid.generation_time,
        "nodeIds": [],
        "recentCursorSearchSelectedNIds": [],
        "recentSearch": [],
        "language": language,
    }
    res = COLL.users.insert_one(data)
    if not res.acknowledged:
        return "", const.Code.OPERATION_FAILED
    un_data: tps.UserNodeIds = {
        "_id": data["_id"],
        "id": data["id"],
        "nodeIds": [],
    }
    res = COLL.unids.insert_one(un_data)
    if not res.acknowledged:
        return "", const.Code.OPERATION_FAILED
    return data["id"], const.Code.OK


def update(
        uid: str,
        email: str = "",
        hashed: str = "",
        nickname: str = "",
        avatar: str = "",
        language: str = "",
) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u, code = get(uid=uid)
    if code != const.Code.OK:
        return None, code
    email = email.strip()
    if email not in ["", u["email"]] and COLL.users.find_one(
            {"account": email, "source": const.UserSource.EMAIL.value}
    ) is not None:
        return None, const.Code.EMAIL_OCCUPIED
    if email == "":
        email = u["email"]
    hashed = hashed.strip()
    if hashed == "":
        hashed = u["hashed"]
    nickname = nickname.strip()
    if nickname == "":
        nickname = u["nickname"]
    avatar = str(avatar).strip()
    if avatar == "":
        avatar = u["avatar"]
    language = language.strip()
    if language == "":
        language = u["language"]
    elif not const.Language.is_valid(language):
        return None, const.Code.INVALID_LANGUAGE

    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {
            "email": email,
            "hashed": hashed,
            "nickname": nickname,
            "avatar": avatar,
            "modifiedAt": datetime.datetime.now(tz=utc),
            "language": language,
        }}
    )
    if res.modified_count != 1:
        return None, const.Code.OPERATION_FAILED
    return get(uid=uid)


def delete(uid: str) -> const.Code:
    res = COLL.users.delete_one({"id": uid})
    return const.Code.OK if res.deleted_count == 1 else const.Code.OPERATION_FAILED


def disable(uid: str) -> const.Code:
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"disabled": True}}
    )
    return const.Code.OK if res.modified_count == 1 else const.Code.OPERATION_FAILED


def enable(uid: str) -> const.Code:
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"disabled": False}}
    )
    return const.Code.OK if res.modified_count == 1 else const.Code.OPERATION_FAILED


def get_by_email(email: str) -> Tuple[Optional[tps.UserMeta], const.Code]:
    if config.get_settings().ONE_USER:
        source = const.UserSource.LOCAL.value
    else:
        source = const.UserSource.EMAIL.value
    return get_account(account=email, source=source)


def get_account(account: str, source: int) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u = COLL.users.find_one({"source": source, "account": account, "disabled": False})
    if u is None:
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    return u, const.Code.OK


def get(uid: str) -> Tuple[Optional[tps.UserMeta], const.Code]:
    u = COLL.users.find_one({"id": uid, "disabled": False})
    if u is None:
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    return u, const.Code.OK


def get_hash_by_uid(uid: str) -> Optional[str]:
    u = COLL.users.find_one({"id": uid, "disabled": False})
    if u is None:
        return None
    return u["hashed"]


def is_exist(uid: str) -> bool:
    try:
        COLL.users.find({"id": uid, "disabled": False}, limit=1).next()
    except StopIteration:
        return False
    return True


def get_node_ids(uid: str) -> Tuple[List[str], const.Code]:
    if not is_exist(uid=uid):
        return [], const.Code.ACCOUNT_OR_PASSWORD_ERROR
    doc = COLL.unids.find_one({"id": uid})
    return doc["nodeIds"], const.Code.OK



import unittest

from rethink.models import utils


class UtilsTest(unittest.TestCase):
    def test_short_uuid(self):
        for _ in range(10):
            uid = utils.short_uuid()
            self.assertEqual(22, len(uid))

    def test_md2txt(self):
        text = utils.md2txt(md="# 123\n## 456\n### 789\n")
        self.assertEqual("123\n456\n789", text)

    def test_pinyin(self):
        res = utils.txt2search_keys('中心English')
        self.assertEqual({'zhongxinenglish', '中心english', 'ㄓㄨㄥㄒㄧㄣenglish'}, set(res.split(" ")))



import os
from pathlib import Path
from typing import Literal, Union

import uvicorn


def run(
        path: Union[str, Path] = None,
        host="127.0.0.1",
        port=8000,
        language: Literal["en", "zh"] = "en",
):
    """
    Run the server.

    Args:
        path: rethink will create .data folder in this path, default is current working directory.
        host: server host, default is the localhost.
        port: server port, default is 8000.
        language: website language, default is English.

    Returns:
        None
    """
    if path is None:
        path = os.getcwd()
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"Path not exists: {path}")
    if not path.is_dir():
        raise NotADirectoryError(f"Path is not a directory: {path}")
    os.environ["LOCAL_STORAGE_PATH"] = str(path)
    os.environ["VUE_APP_API_PORT"] = str(port)
    os.environ["VUE_APP_MODE"] = "local"
    os.environ["VUE_APP_LANGUAGE"] = language
    uvicorn.run(
        "rethink.application:app",
        host=host,
        port=port,
        reload=False,
        workers=1,
        env_file=os.path.join(os.path.abspath(os.path.dirname(__file__)), ".env.local"),
    )



import unittest
from textwrap import dedent

from rethink import const, models
from rethink.models.utils import short_uuid
from . import utils


class LocalModelsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        models.database.drop_all()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def setUp(self) -> None:
        models.database.init()
        u, _ = models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.uid = u["id"]

    def tearDown(self) -> None:
        models.database.drop_all()

    def test_user(self):
        _id, code = models.user.add(
            account="aaa", source=const.UserSource.EMAIL.value,
            email="aaa", hashed="bbb", nickname="ccc", avatar="ddd", language=const.Language.EN.value)
        self.assertNotEqual("", _id)
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("ccc", u["nickname"])

        u, code = models.user.update(uid=_id, email="a@a.com", hashed="1", nickname="2", avatar="3")
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("a@a.com", u["email"])
        self.assertEqual("1", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])

        code = models.user.disable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(uid=_id)
        self.assertEqual(const.Code.ACCOUNT_OR_PASSWORD_ERROR, code)
        self.assertIsNone(u)

        code = models.user.enable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        code = models.user.disable(uid="sdwqdqw")
        self.assertEqual(const.Code.OPERATION_FAILED, code)

        code = models.user.delete(uid=_id)
        self.assertEqual(const.Code.OK, code)

    def test_node(self):
        node, code = models.node.add(
            uid=self.uid, md="a" * (const.MD_MAX_LENGTH + 1), type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.NOTE_EXCEED_MAX_LENGTH, code)
        self.assertIsNone(node)

        node, code = models.node.add(
            uid=self.uid, md="[title](/qqq)\nbody", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("body", n["snippet"])

        ns, total = models.search.user_node(uid=self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        n, code = models.node.update(uid=self.uid, nid=node["id"], md="title2\nbody2")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("title2\nbody2", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        code = models.node.disable(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        code = models.node.to_trash(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)

        code = models.node.delete(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

    def test_parse_at(self):
        nid1, _ = models.node.add(
            uid=self.uid, md="c", type_=const.NodeType.MARKDOWN.value,
        )
        nid2, _ = models.node.add(
            uid=self.uid, md="我133", type_=const.NodeType.MARKDOWN.value,
        )
        md = dedent(f"""title
        fffqw [@c](/n/{nid1['id']})
        fff 
        [@我133](/n/{nid2['id']})
        ffq
        """)
        node, code = models.node.add(
            uid=self.uid, md=md, type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(uid=self.uid)
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)
        found, total = models.search.user_node(uid=self.uid, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(5, total)

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        cache = n["md"]
        n, code = models.node.update(uid=self.uid, nid=node["id"], md=f'{cache}xxxx')
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(cache + "xxxx", n["md"])

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, code = models.node.update(uid=self.uid, nid=node["id"], md=n["title"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    def test_add_set(self):
        node, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.Code.OK, code)

        res = models.db_ops.node_add_to_set(node["id"], "toNodeIds", short_uuid())
        self.assertEqual(1, res.modified_count)
        node, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    def test_cursor_text(self):
        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        q, recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],
            cursor_text="te",
        )
        self.assertEqual(0, len(recom))
        self.assertEqual(None, q)

        q, recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],
            cursor_text="@te",
        )
        self.assertEqual(2, len(recom))
        self.assertEqual("te", q)

        q, recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],  # exclude the second node
            cursor_text="sw21 @te",
        )
        self.assertEqual("te", q)
        self.assertEqual(2, len(recom))
        self.assertEqual("title", recom[1]["title"])

        q, recom = models.node.cursor_query(
            uid=self.uid,
            nid=n2["id"],  # exclude the second node
            cursor_text="sw21 @",
        )
        self.assertEqual("", q)
        self.assertEqual(2, len(recom))

        code = models.search.add_recent_cursor_search(self.uid, n1["id"], n2["id"])
        self.assertEqual(const.Code.OK, code)

        q, recom = models.node.cursor_query(
            uid=self.uid,
            nid=n1["id"],  # exclude the second node
            cursor_text="sw21 @",
        )
        self.assertEqual("", q)
        self.assertEqual(3, len(recom))
        self.assertEqual("Welcome to Rethink", recom[2]["title"])

    def test_to_trash(self):
        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        code = models.node.to_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        ns, total = models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        ns, total = models.search.user_node(self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = models.node.restore_from_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(self.uid)
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    def test_search(self):
        code = models.search.put_recent_search(self.uid, "a")
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        code = models.search.put_recent_search(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        nodes = models.search.get_recent_search(self.uid)
        self.assertEqual(1, len(nodes))
        self.assertEqual(n1["id"], nodes[0]["id"])



from typing import List

from rethink import const, models
from rethink.controllers import schemas
from rethink.controllers.utils import TokenDecode, datetime2str


def __get_node_data(n: models.tps.Node) -> schemas.node.NodeData:
    from_nodes: List[schemas.node.NodeData.LinkedNode] = []
    to_nodes: List[schemas.node.NodeData.LinkedNode] = []
    for nodes, n_nodes in zip(
            [from_nodes, to_nodes],
            [n.get("fromNodes", []), n.get("toNodes", [])]):
        for _n in n_nodes:
            nodes.append(
                schemas.node.NodeData.LinkedNode(
                    id=_n["id"],
                    title=_n["title"],
                    md=_n["md"],
                    snippet=_n["snippet"],
                    type=_n["type"],
                    disabled=_n["disabled"],
                    createdAt=datetime2str(_n["_id"].generation_time),
                    modifiedAt=datetime2str(_n["modifiedAt"]),
                )
            )
    return schemas.node.NodeData(
        id=n["id"],
        md=n["md"],
        title=n["title"],
        type=n["type"],
        disabled=n["disabled"],
        createdAt=datetime2str(n["_id"].generation_time),
        modifiedAt=datetime2str(n["modifiedAt"]),
        fromNodes=from_nodes,
        toNodes=to_nodes,
    )


def put_node(
        td: TokenDecode,
        req: schemas.node.PutRequest,
) -> schemas.node.PutResponse:
    if td.code != const.Code.OK:
        return schemas.node.PutResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            node=None
        )

    n, code = models.node.add(
        uid=td.uid,
        md=req.md,
        type_=req.type,
        from_nid=req.fromNid,
    )
    if code != const.Code.OK:
        return schemas.node.PutResponse(
            code=code.value,
            message=const.get_msg_by_code(code, td.language),
            requestId=req.requestId,
            node=None,
        )
    return schemas.node.PutResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        node=__get_node_data(n),
    )


def get_node(
        td: TokenDecode,
        req_id: str,
        nid: str,
) -> schemas.node.GetResponse:
    n, code = models.node.get(uid=td.uid, nid=nid)
    if code != const.Code.OK:
        return schemas.node.GetResponse(
            requestId=req_id,
            code=code.value,
            message=const.get_msg_by_code(code, td.language),
            node=None,
        )
    return schemas.node.GetResponse(
        requestId=req_id,
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        node=__get_node_data(n),
    )


def update_node(
        td: TokenDecode,
        req: schemas.node.UpdateRequest,
) -> schemas.node.GetResponse:
    if td.code != const.Code.OK:
        return schemas.node.GetResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            node=None
        )
    n, code = models.node.update(
        uid=td.uid,
        nid=req.nid,
        md=req.md,
    )
    return schemas.node.GetResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        node=__get_node_data(n),
    )



import datetime
import unittest

from fastapi.testclient import TestClient
from httpx import Response

from rethink import const
from rethink.application import app
from rethink.models import database
from rethink.models.utils import jwt_decode
from . import utils


class PublicApiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        database.init()

    def setUp(self) -> None:
        self.client = TestClient(app)

    @classmethod
    def tearDownClass(cls) -> None:
        database.drop_all()
        utils.drop_env(".env.test.local")

    def test_home(self):
        resp = self.client.get("/")
        self.assertEqual(200, resp.status_code)
        self.assertEqual("text/html; charset=utf-8", resp.headers["content-type"])
        self.assertEqual(Response, type(resp))

    def test_register(self):
        resp = self.client.put("/api/user", json={
            "email": "a@c.com",
            "password": "a",
            "language": const.Language.EN.value,
            "requestId": "xxx"
        })
        rj = resp.json()
        self.assertEqual(const.Code.ONE_USER_MODE.value, rj["code"])
        self.assertEqual(len(rj["token"]), 0)
        self.assertEqual("xxx", rj["requestId"])


class TokenApiTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    def setUp(self) -> None:
        database.init()
        self.client = TestClient(app)
        resp = self.client.post("/api/login", json={
            "email": const.DEFAULT_USER["email"],
            "password": "",
        })
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.token = rj["token"]

    def tearDown(self) -> None:
        database.drop_all()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    def test_get_user(self):
        resp = self.client.get(
            "/api/user",
            params={"rid": "xxx"},
            headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("rethink", rj["user"]["nickname"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertGreater(
            datetime.datetime.strptime(rj["user"]["createdAt"], "%Y-%m-%dT%H:%M:%SZ").timestamp(),
            0., msg=rj["user"]["createdAt"])

    def test_update_user(self):
        resp = self.client.post("/api/user", json={
            "requestId": "xxx",
            "nickname": "new nickname",
            "avatar": "http://new.avatar/aa.png",
        }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])

        resp = self.client.get("/api/user", params={"rid": "xxx"}, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("new nickname", rj["user"]["nickname"])
        self.assertEqual("http://new.avatar/aa.png", rj["user"]["avatar"])
        self.assertEqual("xxx", rj["requestId"])

    def test_recent_search(self):
        resp = self.client.put("/api/node", json={
            "requestId": "xxx",
            "md": "node1\ntext",
            "type": const.NodeType.MARKDOWN.value,
        }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.put(
            "/api/search/recent",
            json={
                "requestId": "xxx",
                "nid": rj["node"]["id"],
            }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.get(
            "/api/search/recent",
            params={"rid": "xxx"},
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual(len(rj["nodes"]), 1)
        self.assertEqual(rj["nodes"][0]["title"], "node1")

    def test_node(self):
        resp = self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})
        rj = resp.json()
        self.assertGreater(len(rj["data"]["nodes"]), 0)

        self.client.post(
            "/api/search/node",
            json={
                "query": "",
                "requestId": "xxx",
                "sortKey": "createdAt",
                "sortOrder": -1, "page": 0, "pageSize": 5},
            headers={"token": self.token})

        resp = self.client.put("/api/node", json={
            "requestId": "xxx",
            "md": "node1\ntext",
            "type": const.NodeType.MARKDOWN.value,
        }, headers={"token": self.token})
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        node = rj["node"]

        resp = self.client.get(
            "/api/node",
            params={"rid": "xxx", "nid": node["id"]},
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        n = rj["node"]
        self.assertEqual("node1", n["title"], msg=rj)
        self.assertEqual("node1\ntext", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        resp = self.client.post(
            "/api/node",
            json={
                "requestId": "xxx",
                "nid": node["id"],
                "md": "node2\ntext"
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])

        resp = self.client.get(
            "/api/node",
            params={"rid": "xxx", "nid": node["id"]},
            headers={"token": self.token}
        )
        rj = resp.json()
        n = rj["node"]
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual("node2", n["title"])
        self.assertEqual("node2\ntext", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        resp = self.client.put(
            "/api/trash",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.get(
            "/api/trash",
            params={"rid": "xxx", "p": 0, "ps": 10},
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual(1, len(rj["data"]["nodes"]))

        resp = self.client.post(
            "/api/trashRestore",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.post(
            "/api/search/cursor",
            json={
                "requestId": "xxx",
                "nid": node["id"],
                "textBeforeCursor": "@How",
            },
            headers={"token": self.token},
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])
        self.assertEqual("How", rj["result"]["query"])
        self.assertEqual(1, len(rj["result"]["nodes"]))
        self.assertEqual("How do I record", rj["result"]["nodes"][0]["title"])

        self.client.put(
            "/api/trash",
            json={
                "requestId": "xxx",
                "nid": node["id"],
            },
            headers={"token": self.token}
        )
        resp = self.client.delete(
            f"/api/trash/{node['id']}",
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])

        resp = self.client.delete(
            "/api/trash/ssa",
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(const.Code.NODE_NOT_EXIST.value, rj["code"])

        resp = self.client.get(
            "/api/node",
            params={"rid": "xxx", "nid": node["id"]},
            headers={"token": self.token}
        )
        rj = resp.json()
        self.assertEqual(const.Code.NODE_NOT_EXIST.value, rj["code"])

    def test_captcha(self):
        resp = self.client.get(
            "/api/captcha/img",
        )
        self.assertEqual(200, resp.status_code)
        self.assertEqual("image/png", resp.headers["content-type"])
        self.assertGreater(len(resp.headers["x-captcha-token"]), 0)

        code_str = jwt_decode(resp.headers["x-captcha-token"])["code"]
        resp = self.client.post(
            "/api/captcha/img",
            json={
                "requestId": "xxx",
                "token": resp.headers["x-captcha-token"],
                "codeStr": code_str,
            }
        )
        rj = resp.json()
        self.assertEqual(0, rj["code"])
        self.assertEqual("xxx", rj["requestId"])



from rethink import const, models
from rethink.controllers import schemas
from rethink.controllers.utils import TokenDecode


def move_to_trash(
        td: TokenDecode,
        req: schemas.node.RestoreFromTrashRequest,
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
        )
    code = models.node.to_trash(uid=td.uid, nid=req.nid)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
    )


def get_from_trash(
        td: TokenDecode,
        p: int = 0,
        ps: int = 10,
        rid: str = "",
) -> schemas.node.GetFromTrashResponse:
    ns, total = models.node.get_nodes_in_trash(uid=td.uid, page=p, page_size=ps)
    code = const.Code.OK
    return schemas.node.GetFromTrashResponse(
        requestId=rid,
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        data=schemas.node.parse_nodes_info(ns, total)
    )


def restore_from_trash(
        td: TokenDecode,
        req: schemas.node.RestoreFromTrashRequest,
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
        )
    code = models.node.restore_from_trash(uid=td.uid, nid=req.nid)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
    )


def delete_node(
        td: TokenDecode,
        nid: str,
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId="",
        )
    code = models.node.delete(uid=td.uid, nid=nid)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId="",
    )



import os

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles

from rethink import const
from rethink.logger import logger
from .models import database
from .routes import user, oauth, node, search, trash, verification

app = FastAPI(
    docs_url="/docs",
    openapi_url="/openapi.json",
)

if os.getenv("VUE_APP_MODE", "local") not in ["development", "local"]:
    allow_origins = [
        "https://rethink.run",
        "https://www.rethink.run",
    ]
else:
    allow_origins = [
        "*",
    ]

app.add_middleware(
    CORSMiddleware,
    allow_origins=allow_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["X-Captcha-Token"],
)

app.include_router(node.router)
app.include_router(user.router)
app.include_router(oauth.router)
app.include_router(search.router)
app.include_router(trash.router)
app.include_router(verification.router)


@app.on_event("startup")
async def startup_event():
    database.init()
    logger.info("db initialized")


app.mount(
    "/css",
    StaticFiles(directory=const.FRONTEND_DIR / "css"),
    name="css",
)
app.mount(
    "/js",
    StaticFiles(directory=const.FRONTEND_DIR / "js"),
    name="js",
)
app.mount(
    "/img",
    StaticFiles(directory=const.FRONTEND_DIR / "img"),
    name="img",
)


@app.get("/", response_class=HTMLResponse)
@app.get("/r", response_class=HTMLResponse)
@app.get("/user", response_class=HTMLResponse)
@app.get("/login", response_class=HTMLResponse)
async def index() -> HTMLResponse:
    content = (const.FRONTEND_DIR / "index.html").read_text(encoding="utf-8")
    content += f"<script>window.VUE_APP_API_PORT = {os.getenv('VUE_APP_API_PORT')}</script>"
    return HTMLResponse(
        content=content,
        status_code=200,
    )



import datetime
import os
from dataclasses import dataclass
from typing import Optional, Union

from bson import ObjectId
from bson.tz_util import utc
from mongita import MongitaClientDisk
from mongita.collection import Collection
from pymongo import MongoClient
from pymongo.collection import Collection as RemoteCollection

from rethink import config, const
from rethink.logger import logger
from . import utils
from .tps import UserMeta, UserNodeIds, Node


@dataclass
class Collections:
    users: Union[Collection, RemoteCollection] = None
    unids: Union[Collection, RemoteCollection] = None
    nodes: Union[Collection, RemoteCollection] = None


COLL = Collections()
CLIENT: Optional[Union[MongoClient, MongitaClientDisk]] = None


def set_client():
    global CLIENT
    conf = config.get_settings()
    if config.is_local_db():
        if not os.path.exists(conf.LOCAL_STORAGE_PATH):
            raise FileNotFoundError(f"Path not exists: {conf.LOCAL_STORAGE_PATH}")
        dot_data_path = os.path.join(conf.LOCAL_STORAGE_PATH, ".data")
        os.makedirs(dot_data_path, exist_ok=True)
        CLIENT = MongitaClientDisk(dot_data_path)
    else:
        CLIENT = MongoClient(
            host=conf.DB_HOST,
            port=conf.DB_PORT,
            username=conf.DB_USER,
            password=conf.DB_PASSWORD,
        )


def init():
    set_client()
    db = CLIENT[config.get_settings().DB_NAME]
    COLL.users = db["users"]
    COLL.unids = db["unids"]
    COLL.nodes = db["nodes"]

    if config.is_local_db():
        # try fix TypeError: can't compare offset-naive and offset-aware datetimes
        docs = COLL.nodes.find()
        for doc in docs:
            COLL.nodes.update_one(
                {"_id": doc["_id"]},
                {"$set": {"modifiedAt": doc["modifiedAt"].replace(tzinfo=utc)}},
            )

    if isinstance(CLIENT, MongoClient):
        # try creating index
        COLL.users.create_index("id", unique=True)
        COLL.users.create_index(["account", "source"], unique=True)
        COLL.nodes.create_index("id", unique=True)
        COLL.unids.create_index("id", unique=True)

    # try add default user
    if config.get_settings().ONE_USER:
        if COLL.users.find_one(
                {"account": const.DEFAULT_USER["email"], "source": const.UserSource.EMAIL.value}
        ) is not None:
            return

        logger.info("running at the first time, a user with initial data will be created")
        language = os.getenv("VUE_APP_LANGUAGE", const.Language.EN.value)
        ns = const.NEW_USER_DEFAULT_NODES[language]

        def create_node(md: str):
            title_, snippet_ = utils.preprocess_md(md)
            n: Node = {
                "_id": ObjectId(),
                "id": utils.short_uuid(),
                "title": title_,
                "snippet": snippet_,
                "md": md,
                "type": const.NodeType.MARKDOWN.value,
                "disabled": False,
                "inTrash": False,
                "modifiedAt": datetime.datetime.now(tz=utc),
                "inTrashAt": None,
                "fromNodeIds": [],
                "toNodeIds": [],
                "searchKeys": utils.txt2search_keys(title_),
            }
            res = COLL.nodes.insert_one(n)
            if not res.acknowledged:
                raise ValueError("cannot insert default node")
            return n

        n0 = create_node(ns[0])
        _md = ns[1].format(n0["id"])
        n1 = create_node(_md)

        u: UserMeta = {
            "_id": ObjectId(),
            "id": utils.short_uuid(),
            "source": const.UserSource.LOCAL.value,
            "account": const.DEFAULT_USER["email"],
            "email": const.DEFAULT_USER["email"],
            "hashed": "",
            "avatar": const.DEFAULT_USER["avatar"],
            "disabled": False,
            "nickname": const.DEFAULT_USER["nickname"],
            "modifiedAt": datetime.datetime.now(tz=utc),
            "recentCursorSearchSelectedNIds": [n0["id"], n1["id"]],
            "recentSearch": [],
            "language": language,
        }
        _ = COLL.users.insert_one(u)
        unids: UserNodeIds = {
            "_id": u["_id"],
            "id": u["id"],
            "nodeIds": [n0["id"], n1["id"]],
        }
        _ = COLL.unids.insert_one(unids)


def get_client():
    return CLIENT


def drop_all():
    set_client()
    for db_name in CLIENT.list_database_names():
        if db_name == "admin":
            continue
        CLIENT.drop_database(db_name)



from fastapi import Depends, APIRouter
from typing_extensions import Annotated

from rethink.controllers import schemas
from rethink.controllers.auth import token2uid
from rethink.controllers.node import node_ops
from rethink.controllers.utils import TokenDecode
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["node"],
    responses={404: {"description": "Not found"}},
)


@router.put(
    path="/node",
    response_model=schemas.node.PutResponse,
)
@measure_time_spend
async def put_node(
        req: schemas.node.PutRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.node.PutResponse:
    return node_ops.put_node(
        td=token_decode,
        req=req,
    )


@router.get(
    path="/node",
    response_model=schemas.node.GetResponse,
)
@measure_time_spend
async def get_node(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        nid: str,
        rid: str = "",
) -> schemas.node.GetResponse:
    return node_ops.get_node(
        td=token_decode,
        req_id=rid,
        nid=nid,
    )


@router.post(
    path="/node",
    response_model=schemas.node.GetResponse,
)
@measure_time_spend
async def update_node(
        req: schemas.node.UpdateRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.node.GetResponse:
    return node_ops.update_node(
        td=token_decode,
        req=req,
    )



from fastapi import Depends, APIRouter
from typing_extensions import Annotated

from rethink.controllers import schemas
from rethink.controllers.auth import token2uid
from rethink.controllers.node import trash_ops
from rethink.controllers.utils import TokenDecode
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["trash"],
    responses={404: {"description": "Not found"}},
)


@router.put(
    path="/trash",
    response_model=schemas.base.AcknowledgeResponse,
)
@measure_time_spend
async def move_to_trash(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        req: schemas.node.RestoreFromTrashRequest
) -> schemas.base.AcknowledgeResponse:
    return trash_ops.move_to_trash(
        td=token_decode,
        req=req
    )


@router.get(
    path="/trash",
    response_model=schemas.node.GetFromTrashResponse,
)
@measure_time_spend
async def get_from_trash(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        p: int = 0,
        ps: int = 10,
) -> schemas.node.GetFromTrashResponse:
    return trash_ops.get_from_trash(
        td=token_decode,
        p=p,
        ps=ps,
    )


@router.post(
    path="/trashRestore",
    response_model=schemas.base.AcknowledgeResponse,
)
@measure_time_spend
async def restore_node_in_trash(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        req: schemas.node.RestoreFromTrashRequest,
) -> schemas.base.AcknowledgeResponse:
    return trash_ops.restore_from_trash(
        td=token_decode,
        req=req,
    )


@router.delete(
    path="/trash/{nid}",
    response_model=schemas.base.AcknowledgeResponse,
)
@measure_time_spend
async def delete_node(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        nid: str,
) -> schemas.base.AcknowledgeResponse:
    return trash_ops.delete_node(
        td=token_decode,
        nid=nid,
    )



from typing import Any, Dict, Optional

from mongita.results import UpdateResult

from rethink import config
from .database import COLL
from .tps import Node


def remove_from_node(from_nid: str, to_nid: str):
    if config.is_local_db():
        to_n = COLL.nodes.find_one({"id": to_nid})
        if to_n is None:
            return
        try:
            to_n["fromNodeIds"].remove(from_nid)
            COLL.nodes.update_one(
                {"id": to_nid},
                {"$set": {"fromNodeIds": to_n["fromNodeIds"]}}
            )
        except ValueError:
            pass
    else:
        COLL.nodes.update_one(
            {"id": to_nid},
            {"$pull": {"fromNodeIds": from_nid}}
        )


# def remove_to_node(from_nid: str, to_nid: str):
#     if config.is_local_db():
#         from_n = COLL.nodes.find_one({"id": from_nid})
#         if from_n is None:
#             return
#         try:
#             from_n["toNodeIds"].remove(to_nid)
#             COLL.nodes.update_one(
#                 {"id": from_nid},
#                 {"$set": {"toNodeIds": from_n["toNodeIds"]}}
#             )
#         except ValueError:
#             pass
#     else:
#         COLL.nodes.update_one(
#             {"id": from_nid},
#             {"$pull": {"toNodeIds": to_nid}}
#         )


def node_add_to_set(id_: str, key: str, value: Any) -> UpdateResult:
    res = UpdateResult(0, 0)
    if config.is_local_db():
        has_new = False
        doc = COLL.nodes.find_one({"id": id_})
        if doc is None:
            return res
        if key not in doc:
            doc[key] = []
        if value not in doc[key]:
            doc[key].append(value)
            has_new = True
        if has_new:
            res = COLL.nodes.update_one(
                {"id": id_},
                {"$set": {key: doc[key]}}
            )
    else:
        res = COLL.nodes.update_one(
            {"id": id_},
            {"$addToSet": {key: value}}
        )
    return res


def node_get(
        id_: str,
        assert_conditions: Dict[str, Any]
) -> Optional[Node]:
    c = {"id": id_}
    if not config.is_local_db():
        c.update(assert_conditions)

    doc = COLL.nodes.find_one(c)
    if doc is None:
        return None
    if config.is_local_db():
        for k, c in assert_conditions.items():
            if doc[k] != c:
                return None
    return doc


def unids_pull(id_: str, key: str, value: Any) -> UpdateResult:
    res = UpdateResult(0, 0)
    if config.is_local_db():
        doc = COLL.unids.find_one({"id": id_})
        if doc is None:
            return res
        if key not in doc:
            doc[key] = []
        if value in doc[key]:
            doc[key].remove(value)
            res = COLL.unids.update_one(
                {"id": id_},
                {"$set": {key: doc[key]}}
            )
    else:
        res = COLL.users.update_one(
            {"id": id_},
            {"$pull": {key: value}}
        )
    return res



from typing import List, Tuple, Sequence

from rethink import const, config
from . import user, tps
from .database import COLL


def user_node(
        uid: str,
        query: str = "",
        sort_key: str = "createAt",
        sort_order: int = -1,
        page: int = 0,
        page_size: int = 0,
        nid_exclude: Sequence[str] = None,
) -> Tuple[List[tps.Node], int]:
    unids, code = user.get_node_ids(uid=uid)
    if code != const.Code.OK:
        return [], 0

    if nid_exclude is None or len(nid_exclude) == 0:
        nids = unids
    else:
        nid_exclude = nid_exclude or []
        nid_exclude = set(nid_exclude)
        nids = set(unids)
        nids.difference_update(nid_exclude)
        nids = list(nids)

    condition = {
        "id": {"$in": nids},
        "disabled": False,
        "inTrash": False,
    }
    total = COLL.nodes.count_documents(condition)

    query = query.strip().lower()

    # on remote mongodb
    if query != "" and not config.is_local_db():
        condition["$or"] = [
            {"searchKeys": {"$regex": query}},
            {"md": {"$regex": query}},
        ]

    docs = COLL.nodes.find(condition)

    if sort_key != "":
        if sort_key == "createAt":
            sort_key = "_id"
        elif sort_key == "similarity":
            sort_key = "_id"  # TODO: sort by similarity
        docs = docs.sort(sort_key, direction=sort_order)

    if page_size > 0:
        docs = docs.skip(page * page_size).limit(page_size)

    if config.is_local_db() and query != "":
        return [doc for doc in docs if query in doc["searchKeys"] or query in doc["md"]], total
    return list(docs), total


def add_recent_cursor_search(
        uid: str,
        nid: str,
        to_nid: str,
) -> const.Code:
    # add selected node to recentCursorSearchSelectedNIds
    user_c = {"id": uid}
    unid_c = {"id": uid}

    # on remote mongodb
    if not config.is_local_db():
        user_c.update({"disabled": False})
        unid_c.update({"nodeIds": {"$in": [nid, to_nid]}})

    # try finding user
    u = COLL.users.find_one(user_c)
    if u is None:
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR

    # try finding node
    unids = COLL.unids.find_one(unid_c)
    if unids is None:
        return const.Code.NODE_NOT_EXIST

    # do it on local db
    if config.is_local_db():
        if u["disabled"]:
            return const.Code.ACCOUNT_OR_PASSWORD_ERROR
        if nid not in unids["nodeIds"] or to_nid not in unids["nodeIds"]:
            return const.Code.NODE_NOT_EXIST

    rns = u["recentCursorSearchSelectedNIds"]
    if to_nid in rns:
        rns.remove(to_nid)
    rns.insert(0, to_nid)
    if len(rns) > 10:
        rns = rns[:10]

    # add to recentCursorSearchSelectedNIds
    res = COLL.users.update_one(
        {"id": uid},
        {"$set": {"recentCursorSearchSelectedNIds": rns}}
    )
    if res.matched_count != 1:
        return const.Code.OPERATION_FAILED

    return const.Code.OK


def get_recent_search(uid: str) -> List[tps.Node]:
    if not user.is_exist(uid=uid):
        return []
    doc = COLL.users.find_one({"id": uid})
    nodes = COLL.nodes.find({"id": {"$in": doc["recentSearch"]}})
    return sorted(list(nodes), key=lambda x: doc["recentSearch"].index(x["id"]))


def put_recent_search(uid: str, nid: str) -> const.Code:
    if not user.is_exist(uid=uid):
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    doc = COLL.users.find_one({"id": uid})
    if COLL.unids.count_documents({"id": uid, "nodeIds": {"$in": [nid]}}) == 0:
        return const.Code.NODE_NOT_EXIST
    rns = doc["recentSearch"]
    try:
        rns.remove(nid)
    except ValueError:
        pass
    rns.insert(0, nid)
    if len(rns) > 20:
        rns = rns[:20]
    _ = COLL.users.update_one(
        {"id": uid},
        {"$set": {"recentSearch": rns}}
    )
    return const.Code.OK



import datetime
import math
import re
import uuid
from io import StringIO
from typing import Tuple

import jwt
import pypinyin
from markdown import Markdown

from rethink import config

HEADERS = {
    'typ': 'jwt',
    'alg': 'RS256'
}
alphabet = "23456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
alphabet_len = len(alphabet)
__padding = int(math.ceil(math.log(2 ** 128, alphabet_len)))
__code_pattern = re.compile(r"^```[^\S\r\n]*[a-z]*?\n(.*?)\n```$", re.MULTILINE | re.DOTALL)


def short_uuid() -> str:
    """
    The output has the most significant digit first.
    """
    number = uuid.uuid4().int
    output = ""

    while number:
        number, digit = divmod(number, alphabet_len)
        output += alphabet[digit]
    if __padding:
        remainder = max(__padding - len(output), 0)
        output = output + alphabet[0] * remainder
    return output[::-1]


def __unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        __unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


# patching Markdown
Markdown.output_formats["plain"] = __unmark_element
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def md2txt(md: str) -> str:
    for found in list(__code_pattern.finditer(md))[::-1]:
        span = found.span()
        code = found.group(1)
        md = f"{md[: span[0]]}{code}{md[span[1]:]}"
    return __md.convert(md)


def preprocess_md(md: str) -> Tuple[str, str]:
    title, body = split_title_body(fulltext=md)
    title = md2txt(title.strip())
    body = body.strip()
    snippet = md2txt(md=body)[:200]
    return title, snippet


def txt2search_keys(txt: str) -> str:
    s1 = "".join(pypinyin.lazy_pinyin(txt))
    s2 = "".join(pypinyin.lazy_pinyin(txt, style=pypinyin.Style.BOPOMOFO))
    s = {s1, s2, txt}
    return " ".join(s).lower()


def split_title_body(fulltext: str) -> (str, str):
    title_body = fulltext.split("\n", maxsplit=1)
    title = title_body[0].strip()
    try:
        body = title_body[1].strip()
    except IndexError:
        body = ""
    return title, body


def jwt_encode(exp_delta: datetime.timedelta, data: dict) -> str:
    """
    Create token

    Args:
        exp_delta: expired delta
        data: payload data

    Returns:
        str: token
    """
    payload = {
        "exp": datetime.datetime.utcnow() + exp_delta
    }
    payload.update(data)
    token = jwt.encode(
        payload=payload,
        key=config.get_settings().JWT_KEY,
        algorithm=HEADERS["alg"],
        headers=HEADERS,
    )
    return token


def jwt_decode(token: str) -> dict:
    return jwt.decode(
        token,
        key=config.get_settings().JWT_KEY_PUB,
        algorithms=[HEADERS["alg"]],
        options={"verify_exp": True}
    )



from dataclasses import dataclass
from enum import Enum, auto, unique
from pathlib import Path
from textwrap import dedent

RETHINK_DIR = Path(__file__).parent
FRONTEND_DIR = RETHINK_DIR / "dist-local"
MD_MAX_LENGTH = 100000


class NodeType(Enum):
    FILE = 0
    MARKDOWN = auto()


class Code(Enum):
    OK = 0
    ACCOUNT_OR_PASSWORD_ERROR = auto()  # 1
    INVALID_AUTH = auto()  # 2
    EXPIRED_AUTH = auto()  # 3
    USER_EXIST = auto()  # 4
    NODE_EXIST = auto()  # 5
    NODE_NOT_EXIST = auto()  # 6
    OPERATION_FAILED = auto()  # 7
    EMAIL_OCCUPIED = auto()  # 8
    EMPTY_CONTENT = auto()  # 10
    INVALID_TITLE = auto()  # 11
    INVALID_LANGUAGE = auto()  # 12
    ONE_USER_MODE = auto()  # 13
    INVALID_PASSWORD = auto()  # 14
    CAPTCHA_ERROR = auto()  # 15
    CAPTCHA_EXPIRED = auto()  # 16
    NOTE_EXCEED_MAX_LENGTH = auto()  # 17


@dataclass
class CodeMessage:
    zh: str
    en: str


CODE_MESSAGES = {
    Code.OK: CodeMessage(zh="成功", en="OK"),
    Code.ACCOUNT_OR_PASSWORD_ERROR: CodeMessage(zh="账号不存在或者密码错误", en="No such user or password error"),
    Code.INVALID_AUTH: CodeMessage(zh="无效的认证信息", en="Invalid authentication information"),
    Code.EXPIRED_AUTH: CodeMessage(zh="认证信息已过期", en="Authentication information has expired"),
    Code.USER_EXIST: CodeMessage(zh="用户已存在", en="User already exists"),
    Code.NODE_EXIST: CodeMessage(zh="节点已存在", en="Node already exists"),
    Code.NODE_NOT_EXIST: CodeMessage(zh="节点不存在", en="Node does not exist"),
    Code.OPERATION_FAILED: CodeMessage(zh="操作失败", en="Operation failed"),
    Code.EMAIL_OCCUPIED: CodeMessage(zh="邮箱已被占用", en="Email is occupied"),
    Code.EMPTY_CONTENT: CodeMessage(zh="内容不能为空", en="Content cannot be empty"),
    Code.INVALID_TITLE: CodeMessage(zh="标题格式错误", en="Title format error"),
    Code.INVALID_LANGUAGE: CodeMessage(zh="无效的语言", en="Invalid language"),
    Code.ONE_USER_MODE: CodeMessage(zh="单用户模式，不支持注册", en="Single user mode, registration is not supported"),
    Code.INVALID_PASSWORD: CodeMessage(zh="密码格式错误", en="Password format error"),
    Code.CAPTCHA_ERROR: CodeMessage(zh="验证码输入错误", en="Captcha not match"),
    Code.CAPTCHA_EXPIRED: CodeMessage(zh="验证码已过期", en="Captcha expired"),
    Code.NOTE_EXCEED_MAX_LENGTH: CodeMessage(zh="内容超过最大长度", en="Content exceed max length"),
}

DEFAULT_USER = {
    "nickname": "rethink",
    "email": "rethink@rethink.run",
    "avatar": "",
}


@unique
class Language(Enum):
    EN = "en"
    ZH = "zh"

    def __str__(self):
        return self.value

    @classmethod
    def from_str(cls, s: str):
        if s == "en":
            return cls.EN
        elif s == "zh":
            return cls.ZH
        else:
            raise ValueError(f"Invalid language: {s}")

    @classmethod
    def is_valid(cls, s: str):
        try:
            cls.from_str(s)
            return True
        except ValueError:
            return False


def get_msg_by_code(code: Code, language: str = Language.EN.value):
    msg = CODE_MESSAGES[code]
    if language == Language.ZH.value:
        return msg.zh
    elif language == Language.EN.value:
        return msg.en
    else:
        raise ValueError(f"Invalid language: {language}")


@unique
class UserSource(Enum):
    TEST = 0
    EMAIL = auto()  # 1
    PHONE = auto()  # 2
    GOOGLE = auto()  # 3
    FACEBOOK = auto()  # 4
    WECHAT = auto()  # 5
    GITHUB = auto()  # 6
    LOCAL = auto()  # 7

    def __str__(self):
        return self.value


NEW_USER_DEFAULT_NODES = {
    Language.EN.value: [
        dedent("""# How do I record
        
        I like to record freely and without any restrictions.
        """),
        dedent("""Welcome to Rethink
        
        Rethink is a knowledge management system. You can take node and manage your them here.

        Use @, you can link any record to create associations. For example [@How do I record](/n/{}) .

        Rethink also supports markdown syntax, allowing you to record richer expressions.
        """),
    ],
    Language.ZH.value: [
        dedent("""# 我如何记录
        
        我喜欢自由自在的记录，不受任何限制。
        """),
        dedent("""欢迎使用 Rethink
        
        Rethink 是一个知识管理系统，你可以在这里记录你的知识，管理你的记录。

        使用 @，你就可以链接任意的记录，创建联想。比如 [@我如何记录](/n/{}) 。

        Rethink 同样也支持 markdown 语法，让你可以记录更丰富的表达。
        """),
    ]
}



import base64
import hashlib
import re
import traceback
from typing import Optional, Tuple

import bcrypt
import jwt
from fastapi import Header

from rethink import config, const, models
from rethink.controllers.utils import TokenDecode
from rethink.logger import logger
from rethink.models.utils import jwt_decode

# at least 6 characters, at most 20 characters, at least one letter and one number
VALID_PASSWORD_PTN = re.compile(r"^(?=.*[A-Za-z])(?=.*\d).{6,20}$")


def token2uid(token: str = Header(...)) -> TokenDecode:
    uid = ""
    err = ""
    language = const.Language.EN.value
    try:
        payload = jwt_decode(token=token)
        u, code = models.user.get(uid=payload["uid"])
        if code != const.Code.OK:
            logger.error(f"models.user.get err: {const.CODE_MESSAGES[code].zh}")
            return TokenDecode(code=code, language=language)
        uid = u["id"]
        language = u["language"]
    except jwt.ExpiredSignatureError:
        code = const.Code.EXPIRED_AUTH
        err = traceback.format_exc().replace("\n", "\\n")
    except jwt.DecodeError:
        code = const.Code.INVALID_AUTH
        err = traceback.format_exc().replace("\n", "\\n")
    except Exception:  # pylint: disable=broad-except
        code = const.Code.INVALID_AUTH
        err = traceback.format_exc().replace("\n", "\\n")
    if code != const.Code.OK:
        logger.error(f"jwt_decode err: {err}")
    return TokenDecode(code=code, uid=uid, language=language)


def get_user_by_email(email: str) -> Tuple[Optional[models.tps.UserMeta], const.Code]:
    return models.user.get_by_email(email=email)


def verify_user(u: models.tps.UserMeta, password: str) -> bool:
    if config.get_settings().ONE_USER:
        return True
    base_pw = _base_password(password=password, email=u["email"])
    db_hash = models.user.get_hash_by_uid(u["id"])
    match = bcrypt.checkpw(base_pw, db_hash.encode("utf-8"))
    return match


def _base_password(password: str, email: str) -> bytes:
    return base64.b64encode(hashlib.sha256(f"{password}&&{email}".encode("utf-8")).digest())


def register_user(
        email: str,
        password: str,
        language: str = const.Language.EN.value,
) -> Tuple[str, const.Code]:
    if config.get_settings().ONE_USER:
        logger.warning("on ONE_USER mode, user registration will be ignored")
        return "", const.Code.ONE_USER_MODE
    if VALID_PASSWORD_PTN.match(password) is None:
        return "", const.Code.INVALID_PASSWORD
    u, code = models.user.get_by_email(email=email)
    if code == const.Code.OK or u is not None:
        return "", const.Code.USER_EXIST

    bpw = _base_password(password=password, email=email)
    hashed = bcrypt.hashpw(bpw, config.get_settings().DB_SALT)
    uid, code = models.user.add(
        account=email,
        source=const.UserSource.EMAIL.value,
        email=email,
        hashed=hashed.decode("utf-8"),
        nickname=email.split("@")[0],
        avatar="",
        language=language,
    )
    if code != const.Code.OK:
        return "", code

    code = models.node.new_user_add_default_nodes(uid=uid, language=language)
    return uid, code



from fastapi import Depends, APIRouter
from typing_extensions import Annotated

from rethink.controllers import schemas
from rethink.controllers.auth import token2uid
from rethink.controllers.user import user_ops
from rethink.controllers.utils import TokenDecode
from rethink.routes.utils import measure_time_spend

router = APIRouter(
    prefix="/api",
    tags=["user"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    "/login",
    response_model=schemas.user.LoginResponse,
)
@measure_time_spend
async def login(
        req: schemas.user.LoginRequest
) -> schemas.user.LoginResponse:
    return user_ops.login(req=req)


@router.put(
    "/user",
    response_model=schemas.user.LoginResponse,
)
@measure_time_spend
async def register(
        req: schemas.user.RegisterRequest
) -> schemas.user.LoginResponse:
    return user_ops.put(req=req)


@router.get(
    path="/user",
    response_model=schemas.user.UserInfoResponse,
)
@measure_time_spend
async def get_user(
        token_decode: Annotated[TokenDecode, Depends(token2uid)],
        rid: str = "",
) -> schemas.user.UserInfoResponse:
    return user_ops.get_user(
        req_id=rid, td=token_decode
    )


@router.post(
    path="/user",
    response_model=schemas.user.UserInfoResponse,
)
@measure_time_spend
async def update_user(
        req: schemas.user.UpdateRequest,
        token_decode: Annotated[TokenDecode, Depends(token2uid)]
) -> schemas.user.UserInfoResponse:
    return user_ops.update_user(
        td=token_decode,
        req=req,
    )



import datetime
import re
from typing import List, Optional, Set, Tuple

from bson import ObjectId
from bson.tz_util import utc

from rethink import config, const
from . import user, tps, utils, db_ops
from .database import COLL
from .search import user_node

AT_PTN = re.compile(r'\[@[\w ]+?]\(([\w/]+?)\)', re.MULTILINE)
CURSOR_AT_PTN = re.compile(r'\s+?@([\w ]*)$')


def __get_linked_nodes(new_md) -> Tuple[set, const.Code]:
    # last first
    cache_current_to_nid: Set[str] = set()

    for match in list(AT_PTN.finditer(new_md))[::-1]:
        l0, l1 = match.span(1)
        link = new_md[l0:l1]
        if link.startswith("/n/"):
            # check existed nodes
            to_nid = link[3:]
            cache_current_to_nid.add(to_nid)
    return cache_current_to_nid, const.Code.OK


def __flush_to_node_ids(
        nid: str,
        orig_to_nid: List[str],
        new_md: str
) -> Tuple[List[str], const.Code]:
    new_to_nid, code = __get_linked_nodes(new_md=new_md)
    if code != const.Code.OK:
        return [], code

    # remove fromNodes for linked nodes
    orig_to_nid = set(orig_to_nid)
    for to_nid in orig_to_nid.difference(new_to_nid):
        db_ops.remove_from_node(from_nid=nid, to_nid=to_nid)

    # add fromNodes for linked nodes
    for to_nid in new_to_nid.difference(orig_to_nid):
        db_ops.node_add_to_set(id_=to_nid, key="fromNodeIds", value=nid)

    return list(new_to_nid), const.Code.OK


def add(
        uid: str,
        md: str,
        type_: int = const.NodeType.MARKDOWN.value,
        from_nid: str = "",
) -> Tuple[Optional[tps.Node], const.Code]:
    if len(md) > const.MD_MAX_LENGTH:
        return None, const.Code.NOTE_EXCEED_MAX_LENGTH
    title, snippet = utils.preprocess_md(md)

    onid = utils.short_uuid()

    from_nids = []
    if from_nid != "":
        from_nids.append(from_nid)
        res = db_ops.node_add_to_set(from_nid, "toNodeIds", onid)
        if res.modified_count != 1:
            return None, const.Code.OPERATION_FAILED

    new_to_node_ids = []
    if md != "":
        new_to_node_ids, code = __flush_to_node_ids(nid=onid, orig_to_nid=[], new_md=md)
        if code != const.Code.OK:
            return None, code
    _id = ObjectId()
    data: tps.Node = {
        "_id": _id,
        "id": onid,
        "title": title,
        "snippet": snippet,
        "md": md,
        "type": type_,
        "disabled": False,
        "inTrash": False,
        "modifiedAt": _id.generation_time,
        "inTrashAt": None,
        "fromNodeIds": from_nids,
        "toNodeIds": new_to_node_ids,
        "searchKeys": utils.txt2search_keys(title),
    }
    res = COLL.nodes.insert_one(data)
    if not res.acknowledged:
        return None, const.Code.OPERATION_FAILED
    res = COLL.unids.update_one(
        {"id": uid},
        {"$push": {"nodeIds": onid}}
    )
    if res.modified_count != 1:
        return None, const.Code.OPERATION_FAILED

    return data, const.Code.OK


def __set_linked_nodes(doc: tps.Node):
    doc["fromNodes"] = list(COLL.nodes.find({"id": {"$in": doc["fromNodeIds"]}}))
    doc["toNodes"] = list(COLL.nodes.find({"id": {"$in": doc["toNodeIds"]}}))


def get(
        uid: str,
        nid: str,
        with_disabled: bool = False,
        in_trash: bool = False,
) -> Tuple[Optional[tps.Node], const.Code]:
    unids, code = user.get_node_ids(uid=uid)
    if code != const.Code.OK:
        return None, code
    if nid not in unids:
        return None, const.Code.NODE_NOT_EXIST
    assert_conditions = {} if with_disabled else {"disabled": False}
    assert_conditions.update({"inTrash": in_trash})
    doc = db_ops.node_get(id_=nid, assert_conditions=assert_conditions)
    if doc is None:
        return None, const.Code.NODE_NOT_EXIST

    __set_linked_nodes(doc=doc)
    return doc, const.Code.OK


def update(
        uid: str,
        nid: str,
        md: str,
) -> Tuple[Optional[tps.Node], const.Code]:
    if len(md) > const.MD_MAX_LENGTH:
        return None, const.Code.NOTE_EXCEED_MAX_LENGTH
    title, snippet = utils.preprocess_md(md)

    if not user.is_exist(uid=uid):
        return None, const.Code.ACCOUNT_OR_PASSWORD_ERROR
    n, code = get(uid=uid, nid=nid)
    if code != const.Code.OK:
        return None, code
    if n["md"] == md:
        return n, const.Code.OK
    new_data = {
        "title": title,
        "searchKeys": utils.txt2search_keys(title),
        "md": md,
        "snippet": snippet,
        "modifiedAt": datetime.datetime.now(tz=utc),
    }
    new_data["toNodeIds"], code = __flush_to_node_ids(
        nid=n["id"], orig_to_nid=n["toNodeIds"], new_md=md)
    if code != const.Code.OK:
        return None, code

    if not config.is_local_db():
        doc = COLL.nodes.find_one_and_update(
            {"id": nid},
            {"$set": new_data},
            return_document=True,  # return updated doc
        )
    else:
        res = COLL.nodes.update_one(
            {"id": nid},
            {"$set": new_data}
        )
        if res.modified_count != 1:
            return None, const.Code.OPERATION_FAILED
        doc = COLL.nodes.find_one({"id": nid})

    if doc is None:
        return None, const.Code.NODE_NOT_EXIST
    __set_linked_nodes(doc=doc)
    return doc, const.Code.OK


def cursor_query(
        uid: str,
        nid: str,
        cursor_text: str,
) -> Tuple[Optional[str], List[tps.Node]]:
    if cursor_text.startswith("@"):
        query = cursor_text[1:].strip()
    else:
        found = CURSOR_AT_PTN.search(cursor_text)
        if found is None:
            return None, []
        query = found.group(1).strip()

    if query == "":
        u = COLL.users.find_one({"id": uid})
        if u is None:
            return query, []
        rn = u["recentCursorSearchSelectedNIds"]
        try:
            rn.remove(nid)
        except ValueError:
            pass
        return query, sorted(list(COLL.nodes.find({"id": {"$in": rn}})), key=lambda x: rn.index(x["id"]))
    nodes_found, _ = user_node(
        uid=uid,
        query=query,
        sort_key="modifiedAt",
        sort_order=-1,
        page=0,
        page_size=8,
        nid_exclude=[nid],
    )
    return query, nodes_found


def to_trash(uid: str, nid: str) -> const.Code:
    n, code = get(uid=uid, nid=nid, with_disabled=True, in_trash=False)
    if code != const.Code.OK:
        return code

    # remove node
    u = COLL.users.find_one({"id": uid})
    try:
        u["recentCursorSearchSelectedNIds"].remove(nid)
    except ValueError:
        pass
    else:
        res = COLL.users.update_one({"id": uid}, {"$set": {
            "recentCursorSearchSelectedNIds": u["recentCursorSearchSelectedNIds"]
        }})
        if res.modified_count != 1:
            return const.Code.OPERATION_FAILED
    res = COLL.nodes.update_one({"id": nid}, {"$set": {
        "inTrash": True,
        "inTrashAt": datetime.datetime.now(tz=utc)
    }})
    if res.modified_count != 1:
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def get_nodes_in_trash(uid: str, page: int, page_size: int) -> Tuple[List[tps.Node], int]:
    unids, code = user.get_node_ids(uid=uid)
    if code != const.Code.OK:
        return [], 0

    condition = {
        "id": {"$in": unids},
        "disabled": False,
        "inTrash": True,
    }
    docs = COLL.nodes.find(condition).sort("inTrashAt", direction=-1)
    total = COLL.nodes.count_documents(condition)
    if page_size > 0:
        docs = docs.skip(page * page_size).limit(page_size)

    return list(docs), total


def restore_from_trash(uid: str, nid: str) -> const.Code:
    n, code = get(uid=uid, nid=nid, with_disabled=False, in_trash=True)
    if code != const.Code.OK:
        return code

    # remove node
    res = COLL.nodes.update_one({"id": nid}, {"$set": {
        "inTrash": False,
        "inTrashAt": None,
    }})
    if res.modified_count != 1:
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def delete(uid: str, nid: str) -> const.Code:
    n, code = get(uid=uid, nid=nid, with_disabled=True, in_trash=True)
    if code != const.Code.OK:
        return code

    # remove fromNodes for linked nodes, not necessary
    # for linked_nid in n["fromNodeIds"]:
    #     db_ops.remove_to_node(from_nid=linked_nid, to_nid=nid)

    # remove toNodes for linked nodes
    for linked_nid in n["toNodeIds"]:
        db_ops.remove_from_node(from_nid=nid, to_nid=linked_nid)

    # remove node
    res = COLL.nodes.delete_one({"id": nid})
    if res.deleted_count != 1:
        return const.Code.OPERATION_FAILED

    # update user nodeIds
    res = db_ops.unids_pull(uid, "nodeIds", nid)
    if res.matched_count != 1:
        return const.Code.OPERATION_FAILED

    return const.Code.OK


def disable(
        uid: str,
        nid: str,
) -> const.Code:
    if not user.is_exist(uid=uid):
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    res = COLL.nodes.update_one(
        {"id": nid},
        {"$set": {"disabled": True}}
    )
    if res.modified_count != 1:
        return const.Code.OPERATION_FAILED
    return const.Code.OK


def new_user_add_default_nodes(language: str, uid: str) -> const.Code:
    lns = const.NEW_USER_DEFAULT_NODES[language]
    n, code = add(
        uid=uid,
        md=lns[0],
    )
    if code != const.Code.OK:
        return code
    _, code = add(
        uid=uid,
        md=lns[1].format(n["id"]),
    )
    return code



import datetime
from functools import lru_cache
from typing import Optional

import bcrypt
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from pydantic import Field, DirectoryPath
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    ONE_USER: bool = Field(default=1, env='ONE_USER')
    LOCAL_STORAGE_PATH: Optional[DirectoryPath] = Field(env='LOCAL_STORAGE_PATH', default=None)
    DB_NAME: str = Field(env='DB_NAME', default="")
    DB_USER: str = Field(env='DB_USER', default="")
    DB_PASSWORD: str = Field(env='DB_PASSWORD', default="")
    DB_HOST: str = Field(env='DB_HOST', default="")
    DB_PORT: int = Field(env='DB_PORT', default=-1)
    DB_SALT: bytes = Field(env='DB_SALT', default=b"")
    JWT_KEY: bytes = Field(env='JWT_KEY', default=b"")
    JWT_KEY_PUB: bytes = Field(env='JWT_KEY_PUB', default=b"")
    JWT_EXPIRED_DAYS: int = Field(default=1, env='JWT_EXPIRED_DAYS')
    JWT_EXPIRED_DELTA: datetime.timedelta = Field(default=datetime.timedelta(days=1), env='JWT_EXPIRED_DELTA')
    OAUTH_REDIRECT_URL: str = Field(env='OAUTH_REDIRECT_URL', default="")
    OAUTH_CLIENT_ID_GITHUB: str = Field(env='OAUTH_CLIENT_ID_GITHUB', default="")
    OAUTH_CLIENT_SEC_GITHUB: str = Field(env='OAUTH_CLIENT_SEC_GITHUB', default="")
    OAUTH_CLIENT_ID_QQ: str = Field(env='OAUTH_CLIENT_ID_QQ', default="")
    OAUTH_CLIENT_SEC_QQ: str = Field(env='OAUTH_CLIENT_SEC_QQ', default="")
    OAUTH_CLIENT_ID_FACEBOOK: str = Field(env='OAUTH_CLIENT_ID_QQ', default="")
    OAUTH_CLIENT_SEC_FACEBOOK: str = Field(env='OAUTH_CLIENT_SEC_QQ', default="")

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding='utf-8',
        case_sensitive=True,
    )

    def __init__(self):
        super().__init__()
        if self.LOCAL_STORAGE_PATH is None and self.DB_HOST == "":
            raise ValueError("LOCAL_STORAGE_PATH and DB_HOST cannot be empty at the same time")
        if self.DB_SALT == b"":
            self.DB_SALT = bcrypt.gensalt(4)
        if self.JWT_KEY == b"" or self.JWT_KEY_PUB == b"":
            key = rsa.generate_private_key(
                backend=crypto_default_backend(),
                public_exponent=65537,
                key_size=4096
            )
            self.JWT_KEY = key.private_bytes(
                crypto_serialization.Encoding.PEM,
                crypto_serialization.PrivateFormat.PKCS8,
                crypto_serialization.NoEncryption()
            )

            self.JWT_KEY_PUB = key.public_key().public_bytes(
                crypto_serialization.Encoding.OpenSSH,
                crypto_serialization.PublicFormat.OpenSSH
            )
        self.JWT_EXPIRED_DELTA = datetime.timedelta(days=self.JWT_EXPIRED_DAYS)


@lru_cache()
def get_settings() -> Settings:
    return Settings()


def is_local_db() -> bool:
    return get_settings().LOCAL_STORAGE_PATH is not None



import multiprocessing
import shutil
import socket
import time
import unittest
from pathlib import Path

from rethink import run


class TestRun(unittest.TestCase):
    path: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.path = p = Path(__file__).parent / "tmp"
        p.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(str(cls.path), ignore_errors=True)

    def test_run(self):
        port = 8001
        p = multiprocessing.Process(target=run, kwargs={"path": self.path, "port": port, "language": "zh"})
        p.start()
        # p.join()
        while True:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(2)
            result = sock.connect_ex(('127.0.0.1', port))
            sock.close()
            if result == 0:
                break
            time.sleep(0.1)

        p.kill()
        self.assertTrue(self.path.exists())
        self.assertTrue(self.path.is_dir())
        self.assertTrue((self.path / ".data").exists())



import datetime
from typing import List, Optional

from bson import ObjectId
from typing_extensions import TypedDict


class LinkedNode(TypedDict):
    _id: ObjectId
    id: str
    md: str
    title: str
    type: int  # const.NodeType.MARKDOWN.value
    disabled: bool
    modifiedAt: datetime.datetime


class Node(TypedDict):
    _id: ObjectId
    id: str
    md: str
    title: str
    snippet: str
    type: int  # const.NodeType.MARKDOWN.value
    searchKeys: str
    disabled: bool
    inTrash: bool
    modifiedAt: datetime.datetime
    inTrashAt: Optional[datetime.datetime]
    fromNodeIds: List[str]
    toNodeIds: List[str]
    fromNodes: Optional[List[LinkedNode]]
    toNodes: Optional[List[LinkedNode]]


class UserMeta(TypedDict):
    _id: ObjectId
    id: str
    source: int
    account: str
    nickname: str
    email: str
    avatar: str
    hashed: str
    disabled: bool
    modifiedAt: datetime.datetime
    recentSearch: List[str]
    recentCursorSearchSelectedNIds: List[str]
    language: str


class UserNodeIds(TypedDict):
    _id: ObjectId
    id: str
    nodeIds: List[str]



from typing import Optional

from fastapi import Request

from rethink import config, const, models
from rethink.sso.base import SSOLoginError
from rethink.sso.facebook import FacebookSSO
from rethink.sso.github import GithubSSO
from rethink.sso.qq import QQSSO
from .schemas.oauth import OAuthResponse
from .schemas.user import LoginResponse

GITHUB_SSO: Optional[GithubSSO] = None
QQ_SSO: Optional[QQSSO] = None
FACEBOOK_SSO: Optional[FacebookSSO] = None


async def login_github() -> OAuthResponse:
    global GITHUB_SSO
    if GITHUB_SSO is None:
        GITHUB_SSO = GithubSSO(
            client_id=config.get_settings().OAUTH_CLIENT_ID_GITHUB,
            client_secret=config.get_settings().OAUTH_CLIENT_SEC_GITHUB,
            redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/github",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
    uri = await GITHUB_SSO.get_login_url()
    return OAuthResponse(
        code=const.Code.OK.value,
        message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
        uri=uri,
    )


async def callback_github(req: Request) -> LoginResponse:
    try:
        user = await GITHUB_SSO.verify_and_process(req)
    except SSOLoginError:
        code = const.Code.INVALID_AUTH
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    if user is None:
        code = const.Code.INVALID_AUTH
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    u, code = models.user.get_account(account=user.id, source=const.UserSource.GITHUB.value)
    if code == const.Code.OK:
        return LoginResponse(
            requestId="",
            code=const.Code.OK.value,
            message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
            token=models.utils.jwt_encode(
                exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
                data={"uid": u["id"], "language": u["language"]},
            ),
        )

    # no user found, create one
    language = const.Language.EN.value
    uid, code = models.user.add(
        account=user.id,
        source=const.UserSource.GITHUB.value,
        email=user.email if user.email else "",
        hashed="",
        nickname=user.display_name if user.display_name else "",
        avatar=user.picture if user.picture else "",
        language=language,
    )
    if code != const.Code.OK:
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    token = models.utils.jwt_encode(
        exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
        data={"uid": uid, "language": language},
    )
    code = models.node.new_user_add_default_nodes(language=language, uid=uid)

    return LoginResponse(
        requestId="",
        code=code.value,
        message=const.get_msg_by_code(const.Code.OK, language=language),
        token=token,
    )


async def login_facebook() -> OAuthResponse:
    global FACEBOOK_SSO
    if FACEBOOK_SSO is None:
        FACEBOOK_SSO = FacebookSSO(
            client_id=config.get_settings().OAUTH_CLIENT_ID_FACEBOOK,
            client_secret=config.get_settings().OAUTH_CLIENT_SEC_FACEBOOK,
            redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/facebook",
            allow_insecure_http=False,
            use_state=False,
            scope=None,
        )
    uri = await FACEBOOK_SSO.get_login_url()
    return OAuthResponse(
        code=const.Code.OK.value,
        message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
        uri=uri,
    )


async def callback_facebook(req: Request) -> LoginResponse:
    try:
        user = await FACEBOOK_SSO.verify_and_process(req)
    except SSOLoginError:
        code = const.Code.INVALID_AUTH
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    if user is None:
        code = const.Code.INVALID_AUTH
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    u, code = models.user.get_account(account=user.id, source=const.UserSource.GITHUB.value)
    if code == const.Code.OK:
        return LoginResponse(
            requestId="",
            code=const.Code.OK.value,
            message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
            token=models.utils.jwt_encode(
                exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
                data={"uid": u["id"], "language": u["language"]},
            ),
        )

    # no user found, create one
    language = const.Language.EN.value
    uid, code = models.user.add(
        account=user.id,
        source=const.UserSource.FACEBOOK.value,
        email=user.email,
        hashed="",
        nickname=user.display_name,
        avatar=user.picture,
        language=language,
    )
    if code != const.Code.OK:
        return LoginResponse(
            requestId="",
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            token="",
        )
    token = models.utils.jwt_encode(
        exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
        data={"uid": uid, "language": language},
    )
    code = models.node.new_user_add_default_nodes(language=language, uid=uid)
    return LoginResponse(
        requestId="",
        code=code.value,
        message=const.get_msg_by_code(const.Code.OK, language=language),
        token=token,
    )

# async def login_qq() -> OAuthResponse:
#     global QQ_SSO
#     if QQ_SSO is None:
#         QQ_SSO = QQSSO(
#             client_id=config.get_settings().OAUTH_CLIENT_ID_QQ,
#             client_secret=config.get_settings().OAUTH_CLIENT_SEC_QQ,
#             redirect_uri=f"{config.get_settings().OAUTH_REDIRECT_URL}/qq",
#             allow_insecure_http=False,
#             use_state=False,
#             scope=None,
#         )
#     uri = await QQ_SSO.get_login_url()
#     return OAuthResponse(
#         code=const.Code.OK.value,
#         message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
#         uri=uri,
#     )
#
#
# async def callback_qq(req: Request) -> LoginResponse:
#     try:
#         user = await QQ_SSO.verify_and_process(req)
#     except SSOLoginError:
#         code = const.Code.INVALID_AUTH
#         return LoginResponse(
#             requestId="",
#             code=code.value,
#             message=const.get_msg_by_code(code, const.Language.EN.value),
#             token="",
#         )
#     if user is None:
#         code = const.Code.INVALID_AUTH
#         return LoginResponse(
#             requestId="",
#             code=code.value,
#             message=const.get_msg_by_code(code, const.Language.EN.value),
#             token="",
#         )
#     u, code = models.user.get_account(account=user.id, source=const.UserSource.GITHUB.value)
#     if code == const.Code.OK:
#         return LoginResponse(
#             requestId="",
#             code=const.Code.OK.value,
#             message=const.get_msg_by_code(const.Code.OK, const.Language.EN.value),
#             token=controllers.auth.jwt_encode(
#                 uid=u["id"],
#                 language=u["language"],
#             ),
#         )
#
#     # no user found, create one
#     language = const.Language.EN.value
#     uid, code = models.user.add(
#         account=user.id,
#         source=const.UserSource.GITHUB.value,
#         email=user.email,
#         hashed="",
#         nickname=user.display_name,
#         avatar=user.picture,
#         language=language,
#     )
#     if code != const.Code.OK:
#         return LoginResponse(
#             requestId="",
#             code=code.value,
#             message=const.get_msg_by_code(code, const.Language.EN.value),
#             token="",
#         )
#     token = controllers.auth.jwt_encode(
#         uid=uid,
#         language=language,
#     )
#     new_user_node = const.NEW_USER_FIRST_NODE[language]
#     _, code = models.node.add(
#         uid=uid,
#         title=new_user_node["title"],
#         text=new_user_node["text"],
#     )
#     return LoginResponse(
#         requestId="",
#         code=code.value,
#         message=const.get_msg_by_code(const.Code.OK, language=language),
#         token=token,
#     )



import unittest
from textwrap import dedent

from bson import ObjectId

from rethink import const, models, config
from rethink.controllers.auth import register_user
from . import utils


class RemoteModelsTest(unittest.TestCase):
    default_pwd = "rethink123"

    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.development")
        models.database.drop_all()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.development")

    def setUp(self) -> None:
        models.database.init()
        uid, code = register_user(
            email=const.DEFAULT_USER["email"],
            password=self.default_pwd,
            language=const.Language.EN.value)
        self.assertEqual(const.Code.OK, code)
        self.token = models.utils.jwt_encode(
            exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
            data={"uid": uid, "language": const.Language.EN.value},
        )
        self.uid = uid

    def tearDown(self) -> None:
        models.database.drop_all()

    def test_user(self):
        _id, code = register_user(email="aaa", password="bbb123")
        self.assertNotEqual("", str(_id))
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("aaa", u["nickname"])

        u, code = models.user.update(uid=_id, email="a@a.com", hashed="1", nickname="2", avatar="3")
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("a@a.com", u["email"])
        self.assertEqual("1", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])

        code = models.user.disable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        u, code = models.user.get(uid=_id)
        self.assertEqual(const.Code.ACCOUNT_OR_PASSWORD_ERROR, code)

        code = models.user.enable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        code = models.user.disable(uid="ssaa")
        self.assertEqual(const.Code.OPERATION_FAILED, code)

        code = models.user.delete(uid=_id)
        self.assertEqual(const.Code.OK, code)

    def test_node(self):
        node, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title", n["title"])

        ns, total = models.search.user_node(uid=self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        n, code = models.node.update(uid=self.uid, nid=node["id"], md="# title2\ntext2")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("text2", n["snippet"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        code = models.node.disable(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        code = models.node.to_trash(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        code = models.node.delete(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

    def test_parse_at(self):
        nid1, _ = models.node.add(
            uid=self.uid, md="c", type_=const.NodeType.MARKDOWN.value,
        )
        nid2, _ = models.node.add(
            uid=self.uid, md="我133", type_=const.NodeType.MARKDOWN.value,
        )
        md = dedent(
            f"""title
            fffqw [@c](/n/{nid1['id']})
            fff 
            [@我133](/n/{nid2['id']})
            ffq
            """)
        node, code = models.node.add(
            uid=self.uid, md=md, type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(uid=self.uid)
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)

        found, total = models.search.user_node(uid=self.uid, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(5, total)

        n, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        tmp_text = n["md"]
        n, code = models.node.update(uid=self.uid, nid=node["id"], md=tmp_text + "xxxx")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(tmp_text + "xxxx", n["md"])

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, code = models.node.update(uid=self.uid, nid=node["id"], md=n["md"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    def test_add_set(self):
        node, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.Code.OK, code)

        res = models.db_ops.node_add_to_set(node["id"], "toNodeIds", ObjectId())
        self.assertEqual(1, res.modified_count)
        node, code = models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    def test_to_trash(self):
        n1, code = models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        code = models.node.to_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        ns, total = models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        ns, total = models.search.user_node(self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = models.node.restore_from_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)
        nodes, total = models.search.user_node(self.uid)
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)
